class Dict:
    def __init__(self):
        self.dict = {}

    def __bool__(self):
        return True

    def __len__(self):
        return len(self.dict)

    def __setitem__(self, key, value):
        self.dict[key] = value

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.dict[list(self.keys())[key]]
        else:
            return self.dict[key]

    def __getattr__(self, key):
        try:
            return self.__getitem__(key)
        except KeyError:
            return Missing(self, key)

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __contains__(self, key):
        return key in self.dict

    def __iter__(self):
        return iter(self.dict)

    def __dir__(self):
        return super().__dir__() + list(self.dict.keys())

    def __repr__(self):
        class_name = self.__class__.__name__
        args = ", ".join(f"{key!r}" for key in self.keys())
        return f"{class_name}([{args}])"

    def __call__(self, **kwargs):
        self.set(**kwargs)
        return self

    def set(self, **kwargs):
        for key, value in kwargs.items():
            self.dict[key] = value

    def update(self, *args, **kwargs):
        self.dict.update(*args, **kwargs)

    def keys(self):
        return self.dict.keys()

    def values(self):
        return self.dict.values()

    def items(self):
        return self.dict.items()

    def copy(self):
        return self.dict.copy()


class List:
    def __init__(self):
        self.list = []

    def __bool__(self):
        return True

    def __len__(self):
        return len(self.list)

    def __getitem__(self, key):
        return self.list[key]

    def __contains__(self, value):
        return value in self.list

    def __iter__(self):
        return iter(self.list)

    def __repr__(self):
        class_name = self.__class__.__name__
        return f"{class_name}({self.list})"

    def set(self, values):
        self.list = list(values)

    def append(self, value):
        self.list.append(value)

    def copy(self):
        return self.list.copy()


class Missing:
    def __init__(self, obj, key):
        self.obj = obj
        self.key = key

    def __bool__(self):
        return False

    def __getitem__(self, index):
        raise AttributeError(f"{self.obj} has not attribute '{self.key}'")

    def __getattr__(self, index):
        raise AttributeError(f"{self.obj} has not attribute '{self.key}'")



from dataclasses import dataclass
from typing import Callable, Optional

import torch
import torch.utils.data
from torch.optim.lr_scheduler import ReduceLROnPlateau

import ivory.core.trainer
from ivory.core import instance
from ivory.core.run import Run
from ivory.torch import utils

try:
    from apex import amp
except ImportError:
    pass


@dataclass
class Trainer(ivory.core.trainer.Trainer):
    loss: Optional[Callable] = None
    batch_size: int = 32
    shuffle: bool = True
    gpu: bool = False
    precision: int = 32  # Full precision (32), half precision (16).
    amp_level: str = "O1"
    scheduler_step_mode: str = "epoch"

    def __post_init__(self):
        self.loss = instance.get_attr(self.loss)

    def get_dataloader(self, run: Run, mode: str):
        shuffle = self.shuffle if mode == "train" else False
        return torch.utils.data.DataLoader(
            run.datasets[mode], batch_size=self.batch_size, shuffle=shuffle
        )

    def on_fit_begin(self, run):
        if self.gpu:
            run.model.cuda()
            if self.precision == 16:
                run.model, run.optimizer = amp.initialize(
                    run.model, run.optimizer, opt_level=self.amp_level
                )

    def on_train_begin(self, run):
        run.model.train()

    def train_step(self, run, index, input, target):
        if self.gpu:
            input = utils.cuda(input)
            target = utils.cuda(target)
        output = self.forward(run.model, input)
        if run.results:
            run.results.step(index, output, target)
        loss = self.loss(output, target)
        run.metrics.step(loss.item())
        optimizer = run.optimizer
        optimizer.zero_grad()
        if self.gpu and self.precision == 16:
            with amp.scale_loss(loss, optimizer) as scaled_loss:
                scaled_loss.backward()
        else:
            loss.backward()
        optimizer.step()
        if run.sheduler and self.scheduler_step_mode == "batch":
            run.scheduler.step()

    def on_val_begin(self, run):
        run.model.eval()

    @torch.no_grad()
    def val_step(self, run, index, input, target):
        if self.gpu:
            input = utils.cuda(input)
            target = utils.cuda(target)
        output = self.forward(run.model, input)
        if run.results:
            run.results.step(index, output, target)
        loss = self.loss(output, target)
        run.metrics.step(loss.item())

    def on_epoch_end(self, run):
        if run.scheduler and self.scheduler_step_mode == "epoch":
            if isinstance(run.scheduler, ReduceLROnPlateau):
                run.scheduler.step(run.monitor.score)
            else:
                run.scheduler.step()

    def on_test_begin(self, run):
        self.on_fit_begin(run)
        run.model.eval()

    @torch.no_grad()
    def test_step(self, run, index, input, *target):
        if self.gpu:
            input = utils.cuda(input)
        output = self.forward(run.model, input)
        if run.results:
            run.results.step(index, output, *target)

    def forward(self, model, input):
        return model(input)



import ast
from typing import Any, Dict, Iterable, List

from ivory.utils.range import Range


def parse_args(args=None, **kwargs) -> Dict[str, Iterable[Any]]:
    if args is None:
        args = {}
    elif isinstance(args, (list, tuple)):
        args = dict(arg.split("=") for arg in args)
    elif isinstance(args, dict):
        args = args.copy()
    else:
        raise ValueError(f"Invalid arguments type: {type(args)}.")
    args.update(kwargs)
    parsed = {}
    for arg, value in args.items():
        value = parse_value(value)
        if arg.endswith(".log"):
            arg = arg[:-4]
            if not isinstance(value, Range):
                raise ValueError(f"Invalid value for log mode: {value}")
            value.log = True
        parsed[arg] = value
    return parsed


def parse_value(value) -> Iterable[Any]:
    if isinstance(value, (list, tuple)):
        return list(value)
    if not isinstance(value, str):
        return [value]
    if not value:
        return [""]
    if "," in value:
        values: List[Any] = []
        for v in value.split(","):
            values.extend(parse_value(v))
        return values
    try:
        return Range(value)
    except ValueError:
        return [literal_eval(value)]


def literal_eval(value):
    try:
        return ast.literal_eval(value)
    except ValueError:
        return value



import lightgbm as lgb

import ivory.core.estimator
from ivory.core import instance
from ivory.core.run import Run


class Estimator(ivory.core.estimator.Estimator):
    def __init__(self, **kwargs):
        self.params, self.kwargs = instance.filter_params(lgb.train, **kwargs)

    def step(self, run: Run, mode: str):  # type:ignore
        if mode == "train":
            train_dataset = run.datasets.train[:][1:]
            val_dataset = run.datasets.val[:][1:]
            self.fit(train_dataset, val_dataset)
        super().step(run, mode, training=False)

    def fit(self, train_dataset, val_dataset):
        train_set = lgb.Dataset(*train_dataset)
        val_set = lgb.Dataset(*val_dataset)
        valid_sets = [train_set, val_set]
        self.estimator = lgb.train(
            self.params, train_set, valid_sets=valid_sets, **self.kwargs
        )


class Regressor(Estimator):
    def __init__(self, objective="regression", metric="mse", **kwargs):
        super().__init__(objective=objective, metric=metric, **kwargs)


class Classifier(Estimator):
    def __init__(self, objective="multiclass", metric="multi_logloss", **kwargs):
        super().__init__(objective=objective, metric=metric, **kwargs)



def cuda(x):
    if isinstance(x, list):
        return [xi.cuda() for xi in x]
    elif isinstance(x, dict):
        return {key: x[key].cuda() for key in x}
    else:
        return x.cuda()


def cpu(x):
    if isinstance(x, list):
        return [xi.cpu() for xi in x]
    elif isinstance(x, dict):
        return {key: x[key].cpu() for key in x}
    else:
        return x.cpu()



import os
import re
import shutil
import tempfile
from dataclasses import dataclass
from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple, Union

import mlflow
from logzero import logger
from mlflow.tracking.client import MlflowClient
from mlflow.tracking.context import registry as context_registry
from mlflow.tracking.context.git_context import _get_git_commit
from mlflow.utils.mlflow_tags import (MLFLOW_GIT_COMMIT, MLFLOW_PARENT_RUN_ID,
                                      MLFLOW_RUN_NAME, MLFLOW_SOURCE_NAME)

import ivory.core.state
from ivory import utils
from ivory.callbacks.tracking import Tracking
from ivory.core import instance
from ivory.core.run import Run


@dataclass
class Tracker:
    tracking_uri: Optional[str] = None
    artifact_location: Optional[str] = None

    def __post_init__(self):
        self.client = mlflow.tracking.MlflowClient(self.tracking_uri)
        self.tracking_uri = self.client._tracking_client.tracking_uri
        if self.artifact_location:
            self.artifact_location = utils.path.to_uri(self.artifact_location)

    def get_experiment_id(self, name: str):
        experiment = self.client.get_experiment_by_name(name)
        if experiment:
            return experiment.experiment_id

    def get_run_id(self, experiment_id: str, name: str, number: int):
        run_name = create_run_name(name, number)
        for run in self.client.search_runs(experiment_id):
            if get_run_name(run) == run_name:
                return run.info.run_id

    def create_experiment(self, name: str) -> str:
        experiment_id = self.get_experiment_id(name)
        if not experiment_id:
            logger.info(f"A new experiment created with name: {name!r}")
            experiment_id = self.client.create_experiment(name, self.artifact_location)
        return experiment_id

    def get_run_number(self, experiment_id: str, name: str) -> int:
        name = create_run_name(name)
        number = -1
        for run in self.client.search_runs(experiment_id, run_view_type=3):
            run_name = get_run_name(run)
            if run_name.startswith(name):
                number = max(number, int(run_name.split("#")[1]))
        return number

    def create_run(
        self, experiment_id: str, name: str, source_name: str = ""
    ) -> Tuple[str, str]:
        number = self.get_run_number(experiment_id, name) + 1
        run_name = create_run_name(name, number)
        tags = create_tags(run_name, source_name)
        run = self.client.create_run(experiment_id, tags=tags)
        return run_name, run.info.run_id

    def create_tracking(self) -> Tracking:
        return Tracking(self.tracking_uri)  # type:ignore

    def list_experiments(self, view_type=None) -> List:
        return self.client.list_experiments(view_type)

    def list_run_ids(
        self,
        experiment_id: str,
        parent_run_id: Union[str, Iterable[str]] = "",
        exclude_parent: bool = False,
    ) -> Iterator[str]:
        if parent_run_id:
            yield from self.list_nested_run_ids(experiment_id, parent_run_id)
        else:
            if exclude_parent:
                parent_run_ids = list(self.list_parent_run_ids(experiment_id))
            for run_info in self.client.list_run_infos(experiment_id):
                run_id = run_info.run_id
                if not exclude_parent or run_id not in parent_run_ids:
                    yield run_id

    def list_nested_run_ids(
        self, experiment_id: str, parent_run_id: Union[str, Iterable[str]] = ""
    ) -> Iterator[str]:
        if not isinstance(parent_run_id, str):
            for parent_run_id in list(parent_run_id):
                yield from self.list_nested_run_ids(experiment_id, parent_run_id)
            return
        filter_string = ""
        if parent_run_id:
            filter_string = f"tags.{MLFLOW_PARENT_RUN_ID}={parent_run_id!r}"
        for run in self.client.search_runs(experiment_id, filter_string):
            if MLFLOW_PARENT_RUN_ID in run.data.tags:
                yield run.info.run_id

    def list_parent_run_ids(self, experiment_id: str) -> Iterator[str]:
        parent_run_ids: List[str] = []
        for run in self.client.search_runs(experiment_id):
            if MLFLOW_PARENT_RUN_ID in run.data.tags:
                parent_run_id = run.data.tags[MLFLOW_PARENT_RUN_ID]
                if parent_run_id not in parent_run_ids:
                    yield parent_run_id
                    parent_run_ids.append(parent_run_id)

    def remove_deleted_runs(self, experiment_id: str) -> int:
        experiment = self.client.get_experiment(experiment_id)
        uri = experiment.artifact_location
        if not uri.startswith("file"):
            raise NotImplementedError
        path = utils.path.local_file_uri_to_path(uri)
        num_runs = 0
        for run_info in self.client.list_run_infos(experiment_id, run_view_type=2):
            run_id = run_info.run_id
            directory = os.path.normpath(os.path.join(path, run_id))
            if os.path.exists(directory):
                shutil.rmtree(directory)
                num_runs += 1
        return num_runs

    def search_run_ids(
        self,
        experiment_id: str,
        run_name: str = "",
        parent_run_id: Union[str, Iterable[str]] = "",
        parent_only: bool = False,
        nested_only: bool = False,
        exclude_parent: bool = False,
        best_score_limit: Optional[float] = None,
        **query,
    ) -> Iterator[str]:
        if parent_only:
            run_ids = self.list_parent_run_ids(experiment_id)
        elif nested_only:
            run_ids = self.list_nested_run_ids(experiment_id)
        else:
            run_ids = self.list_run_ids(experiment_id, parent_run_id, exclude_parent)
        for run_id in run_ids:
            if run_name and not re.match(run_name, self.get_run_name(run_id)):
                continue
            if query:
                params = self.load_params(run_id)
                if not utils.params.match(params, **query):
                    continue
            if best_score_limit is not None:
                monitor = self.load_instance(run_id, "monitor", "test")
                if monitor.mode == "min" and monitor.best_score <= best_score_limit:
                    yield run_id
                elif monitor.mode == "max" and monitor.best_score >= best_score_limit:
                    yield run_id
            else:
                yield run_id

    def get_run_name(self, run_id: str) -> str:
        return get_run_name(self.client.get_run(run_id))

    def get_run_name_tuple(self, run_id: str) -> Tuple[str, int]:
        run_name = self.get_run_name(run_id)
        return split_run_name(run_name)

    def get_source_name(self, run_id: str) -> str:
        return get_source_name(self.client.get_run(run_id))

    def get_parent_run_id(self, run_id: str) -> str:
        return get_parent_run_id(self.client.get_run(run_id))

    def get_run_status(self, run_id: str) -> str:
        return self.client.get_run(run_id).info.status

    def load_params(self, run_id: str) -> Dict[str, Any]:
        return load(self, run_id, "params")

    def load_run(self, run_id: str, mode: str) -> Run:
        name, _ = self.get_run_name_tuple(run_id)
        run = load(self, run_id, name, mode=mode)
        run.source_name = self.get_source_name(run_id)
        run.set_tracker(self, name)
        return run

    def load_instance(self, run_id: str, instance_name: str, mode: str) -> Any:
        name, _ = self.get_run_name_tuple(run_id)
        return load(self, run_id, name, instance_name, mode=mode)

    def load_state_dict(self, run: Run, mode: str):
        mode = get_valid_mode(self.client, run.id, mode)
        with tempfile.TemporaryDirectory() as tmpdir:
            state_dict_path = self.client.download_artifacts(run.id, mode, tmpdir)
            state_dict = run.load(state_dict_path)
            run.load_state_dict(state_dict)

    def update_params(self, experiment_id: str, **default):
        runs = []
        for run_id in self.list_run_ids(experiment_id, exclude_parent=True):
            runs.append(self.client.get_run(run_id))
        args = []
        for run in runs:
            args.extend(list(run.data.params.keys()))
        args = list(set(args))
        tracking = self.create_tracking()
        for run in runs:
            run_id = run.info.run_id
            params = self.load_params(run_id)
            update = {}
            for arg in args:
                value = utils.params.get_value(params["run"], arg)
                if value is not None:
                    update[arg] = value
                elif arg in default:
                    update[arg] = default[arg]
            tracking.log_params(run_id, update)

    def set_parent_run_id(self, run_id: str, parent_run_id: str):
        self.client.set_tag(run_id, MLFLOW_PARENT_RUN_ID, parent_run_id)


params_cache: Dict[str, Dict[str, Any]] = {}


def load(
    tracker: Tracker, run_id: str, name: str, instance_name: str = "", mode: str = ""
):
    if name == "params" and run_id in params_cache:
        return params_cache[run_id]
    source_name = tracker.get_source_name(run_id)
    client = tracker.client
    with utils.path.chdir(source_name):
        with tempfile.TemporaryDirectory() as tmpdir:
            if run_id not in params_cache:
                params_path = client.download_artifacts(run_id, "params.yaml", tmpdir)
                params, _ = utils.path.load_params(params_path)
                params_cache[run_id] = params
                if name == "params":
                    return params
            params = params_cache[run_id]
            mode = get_valid_mode(client, run_id, mode)
            if not instance_name:
                run = create_run(params, name)
                if mode:
                    state_dict_path = client.download_artifacts(run_id, mode, tmpdir)
                    state_dict = run.load(state_dict_path)
                    run.load_state_dict(state_dict)
                return run
            instance = create_instance(params, name, instance_name)
            if not mode:
                return instance
            os.mkdir(os.path.join(tmpdir, mode))
            path = os.path.join(mode, instance_name)
            state_dict_path = client.download_artifacts(run_id, path, tmpdir)
            if isinstance(instance, ivory.core.state.State):
                state_dict = ivory.core.state.load(state_dict_path)
            else:
                run = create_run(params, name)
                state_dict = run.load_instance(state_dict_path)
            instance.load_state_dict(state_dict)
            return instance


def create_run(params: Dict[str, Any], name: str) -> Run:
    return instance.create_base_instance(params, name)


def create_instance(params: Dict[str, Any], name: str, instance_name: str) -> Any:
    return instance.create_instance(params[name], instance_name)


def get_valid_mode(client: MlflowClient, run_id: str, mode: str) -> str:
    modes = []
    for artifact in client.list_artifacts(run_id):
        if artifact.is_dir:
            modes.append(artifact.path)
    if mode == "test" and mode not in modes:
        mode = "best"
    if mode == "best" and mode not in modes:
        mode = "current"
    if mode == "current" and mode not in modes:
        mode = ""
    return mode


git_cache: Dict[str, str] = {}


def create_run_name(name: str, number: int = -1) -> str:
    # name = name[0].upper() + name[1:]
    if number == -1:
        return name
    # return f"{name}#{number:03d}"
    return f"{name}#{number}"


def split_run_name(run_name: str) -> Tuple[str, int]:
    name, number = run_name.split("#")
    return name.lower(), int(number)


def create_tags(run_name: str, source_name: str = "") -> Dict[str, str]:
    tags = {MLFLOW_RUN_NAME: run_name}
    if source_name:
        tags[MLFLOW_SOURCE_NAME] = source_name
        if source_name not in git_cache:
            git_cache[source_name] = _get_git_commit(source_name)
        if git_cache[source_name]:
            tags[MLFLOW_GIT_COMMIT] = git_cache[source_name]
    tags = context_registry.resolve_tags(tags)
    return tags


def get_run_name(run) -> str:
    return run.data.tags[MLFLOW_RUN_NAME]


def get_source_name(run) -> str:
    return run.data.tags[MLFLOW_SOURCE_NAME]


def get_parent_run_id(run) -> str:
    return run.data.tags[MLFLOW_PARENT_RUN_ID]



"""Metrics to record scores while training."""
from typing import Any, Dict

import ivory.core.collections
from ivory.core import instance
from ivory.core.run import Run
from ivory.core.state import State


class Metrics(ivory.core.collections.Dict, State):
    """Metrics object."""

    def __init__(self, **kwargs):
        super().__init__()
        self.metrics_fn = {}
        for key, value in kwargs.items():
            self.metrics_fn[key] = get_metric_function(key, value)
        self.history = ivory.core.collections.Dict()

    def __str__(self):
        metrics = []
        for metric in self:
            metrics.append(f"{metric}={self[metric]:.4g}")
        return " ".join(metrics)

    def __repr__(self):
        class_name = self.__class__.__name__
        args = str(self).replace(" ", ", ")
        return f"{class_name}({args})"

    def on_epoch_begin(self, run: Run):
        if run.trainer:
            self.epoch = run.trainer.epoch
        else:
            self.epoch = 0

    def on_epoch_end(self, run: Run):
        self.update(self.metrics_dict(run))
        self.update_history()

    def update_history(self):
        for metric, value in self.items():
            if metric not in self.history:
                self.history[metric] = {self.epoch: value}
            else:
                self.history[metric][self.epoch] = value

    def metrics_dict(self, run: Run) -> Dict[str, Any]:
        """Returns an extra custom metrics dictionary."""
        pred = run.results.val.output.reshape(-1)
        true = run.results.val.target.reshape(-1)
        metrics_dict = {}
        for key, func in self.metrics_fn.items():
            metrics_dict[key] = func(true, pred)
        return metrics_dict


METRICS = {"mse": "sklearn.metrics.mean_squared_error"}


def get_metric_function(key, value):
    if value is None:
        if key not in METRICS:
            raise ValueError(f"Unkown metric: {key}")
        value = METRICS[key]
    if isinstance(value, str) and "." not in value:
        value = f"sklearn.metrics.{value}"

    return instance.get_attr(value)



import functools
import gc
import inspect
import os
import warnings
from typing import Any, Dict, Iterable, Iterator, Optional

from termcolor import colored

import ivory.core.collections
import ivory.core.state
from ivory import utils
from ivory.core.base import CallbackCaller
from ivory.utils.tqdm import tqdm


class Run(CallbackCaller):
    def set_tracker(self, tracker, name: str):
        if not self.id:
            self.name, self.id = tracker.create_run(
                self.experiment_id, name, self.source_name
            )
            self.params[name]["name"] = self.name
            self.params[name]["id"] = self.id
        self.set(tracker=tracker)
        self.set(tracking=tracker.create_tracking())

    def init(self, mode: str = "train"):
        self.create_callbacks()
        self.mode = mode
        self.on_init_begin()
        self.on_init_end()

    def start(self, mode: str = "train"):
        if mode == "both":
            self.start("train")
            if self.tracker:
                self.tracker.load_state_dict(self, "best")
                self.start("test")
        else:
            self.init(mode)
            for obj in self.values():
                if hasattr(obj, "start") and callable(obj.start):
                    obj.start(self)

    def state_dict(self):
        state_dict = {}
        for name, obj in self.items():
            if hasattr(obj, "state_dict") and callable(obj.state_dict):
                with warnings.catch_warnings():  # for torch LambdaLR scheduler
                    warnings.simplefilter("ignore")
                    state_dict[name] = obj.state_dict()
        return state_dict

    def load_state_dict(self, state_dict: Dict[str, Any]):
        for name in state_dict:
            with warnings.catch_warnings():  # for torch LambdaLR scheduler
                warnings.simplefilter("ignore")
                self[name].load_state_dict(state_dict[name])

    def save(self, directory: str):
        for name, state_dict in self.state_dict().items():
            path = os.path.join(directory, name)
            if hasattr(self[name], "save") and callable(self[name].save):
                self[name].save(state_dict, path)
            elif isinstance(self[name], ivory.core.state.State):
                ivory.core.state.save(state_dict, path)
            else:
                self.save_instance(state_dict, path)

    def save_instance(self, state_dict: Dict[str, Any], path: str):
        raise NotImplementedError

    def load(self, directory: str) -> Dict[str, Any]:
        state_dict = {}
        for name in os.listdir(directory):
            path = os.path.join(directory, name)
            if hasattr(self[name], "load") and callable(self[name].load):
                state_dict[name] = self[name].load(path)
            elif isinstance(self[name], ivory.core.state.State):
                state_dict[name] = ivory.core.state.load(path)
            else:
                instance_state_dict = self.load_instance(path)
                if instance_state_dict:
                    state_dict[name] = instance_state_dict
        return state_dict

    def load_instance(self, path):
        raise NotImplementedError


class Task(Run):
    def create_run(self, args, **kwargs):
        run = super().create_run(args, **kwargs)
        run_name = colored(f"[{run.name}]", "green")
        msg = utils.params.to_str(args)
        tqdm.write(run_name + f" {msg}")
        if self.tracking:
            self.tracking.set_parent_run_id(run.id, self.id)
        return run

    def terminate(self):
        if self.tracking:
            self.tracking.client.set_terminated(self.id)

    def product(
        self,
        params: Optional[Dict[str, Iterable[Any]]] = None,
        repeat: int = 1,
        **kwargs,
    ) -> Iterator[Run]:
        params, base_params = utils.params.split_params(params, **kwargs)
        if self.tracking:
            self.tracking.set_tags(self.id, params)
            self.tracking.set_tags(self.id, base_params)
        params_list = list(utils.params.product(params)) * repeat
        if "verbose" not in base_params or base_params["verbose"]:
            params_list = tqdm(params_list, desc="Prod ")
        for args_ in params_list:
            args = base_params.copy()
            args.update(args_)
            run = self.create_run(args)
            yield run
            del run
            gc.collect()
        self.terminate()

    def chain(
        self,
        params: Optional[Dict[str, Iterable[Any]]] = None,
        use_best_param: bool = True,
        **kwargs,
    ) -> Iterator[Run]:
        params, base_params = utils.params.split_params(params, **kwargs)
        if self.tracking:
            self.tracking.set_tags(self.id, params)
            self.tracking.set_tags(self.id, base_params)
        params_list = {arg: list(value) for arg, value in params.items()}
        total = sum(len(value) for value in params_list.values())
        if "verbose" in base_params and base_params["verbose"] == 0:
            bar = None
        else:
            bar = tqdm(total=total, desc="Chain")
        best_params: Dict[str, Any] = {}
        for arg, values in params_list.items():
            best_param = None
            for value in values:
                args = base_params.copy()
                args.update({arg: value})
                if use_best_param:
                    args.update(best_params)
                run = self.create_run(args)
                yield run
                if run.monitor:
                    current_score = run.monitor.best_score
                    if best_param is None:
                        best_score = current_score
                        best_param = value
                    elif run.monitor.mode == "min" and current_score < best_score:
                        best_score = current_score
                        best_param = value
                    if current_score > best_score:
                        best_score = current_score
                        best_param = value
                del run
                gc.collect()
            if best_param is not None:
                best_params[arg] = best_param
            if bar is not None:
                bar.update(1)
        self.terminate()


class Study(Task):
    def optimize(self, suggest_name: str = "", **kwargs):
        if not suggest_name:
            suggest_name = list(self.objective.suggests.keys())[0]
        study_name = ".".join([self.experiment_name, suggest_name, self.name])
        mode = self.create_instance("monitor").mode
        study = self.tuner.create_study(study_name, mode)
        if self.tracking:
            self.tracking.set_tags(self.id, {"study_name": study_name})
            study.set_user_attr("run_id", self.id)
        has_pruning = self.tuner.pruner is not None
        optimize_args = inspect.signature(study.optimize).parameters.keys()
        params = {}
        for key in list(kwargs.keys()):
            if key not in optimize_args:
                params[key] = kwargs.pop(key)
        create_run = functools.partial(self.create_run, **params)
        objective = self.objective(suggest_name, create_run, has_pruning)
        study.optimize(objective, **kwargs)
        self.terminate()
        return study



import ast
import contextlib
import os
import urllib.parse
import urllib.request
from typing import Any, Dict, Tuple

import yaml

Params = Dict[str, Any]


def to_uri(path: str) -> str:
    if urllib.parse.urlparse(path).scheme:
        return path
    if "~" in path:
        path = os.path.expanduser(path)
    url = os.path.abspath(path)
    if "\\" in url:
        url = urllib.request.pathname2url(path)
    return urllib.parse.urlunparse(("file", "", url, "", "", ""))


def local_file_uri_to_path(uri):
    """
    Convert URI to local filesystem path.
    No-op if the uri does not have the expected scheme.
    """
    path = urllib.parse.urlparse(uri).path if uri.startswith("file:") else uri
    return urllib.request.url2pathname(path)


@contextlib.contextmanager
def chdir(source_name: str):
    curdir = os.getcwd()
    if source_name:
        basedir = os.path.dirname(source_name)
        os.chdir(basedir)
    try:
        yield
    finally:
        os.chdir(curdir)


def normpath(name: str, directory: str = "") -> str:
    """Returns the absolute path with the extension."""
    path = os.path.abspath(os.path.join(directory, name))
    if os.path.exists(path + ".yaml"):
        return path + ".yaml"
    if os.path.exists(path + ".yml"):
        return path + ".yml"
    else:
        return path


def load_params(name: str, source_name: str = "") -> Tuple[Params, str]:
    if source_name:
        directory = os.path.dirname(source_name)
        source_name = normpath(name, directory)
    else:
        source_name = name
    with open(source_name, "r") as file:
        params_yaml = file.read()
    params = yaml.safe_load(params_yaml)
    params = literal_eval(params)
    update_include(params, source_name)
    params = inherit(params, source_name)
    return params, source_name


def update_include(params, source_name, include=None):
    if "include" in params:
        name = params.pop("include")
        include = load_params(name, source_name)[0]
    elif include is None:
        include = {}
    for key, value in params.items():
        if key in include:
            if value is None:
                params[key] = include[key]
            elif isinstance(value, dict):
                for k in include[key]:
                    if k not in value:
                        value[k] = include[key][k]
        if isinstance(value, dict):
            update_include(value, source_name, include)


def inherit(params, source_name):
    if "base" in params:
        return _inherit(params, source_name)
    for key, value in params.items():
        if isinstance(value, dict):
            params[key] = inherit(value, source_name)
    return params


def _inherit(params, source_name):
    name = params.pop("base")
    base = load_params(name, source_name)[0]
    for key, value in base.items():
        if key in params:
            if value is None:
                base[key] = params[key]
            elif isinstance(value, dict):
                for k in params[key]:
                    value[k] = params[key][k]
    return base


def literal_eval(x):
    if isinstance(x, dict):
        return {key: literal_eval(value) for key, value in x.items()}
    elif isinstance(x, list):
        return [literal_eval(value) for value in x]
    elif isinstance(x, str):
        try:
            v = ast.literal_eval(x)
        except Exception:
            return x
        if isinstance(v, int):
            return x
        if isinstance(v, float) and "e" not in x and "E" not in x:
            return x
        return v
    else:
        return x



from dataclasses import dataclass

import numpy as np

from ivory.core.run import Run
from ivory.core.state import State


@dataclass
class Monitor(State):
    metric: str = "val_loss"
    mode: str = "min"

    def __post_init__(self):
        self.is_best = False
        self.best_epoch = -1
        if self.mode == "min":
            self.best_score = np.inf
        elif self.mode == "max":
            self.best_score = -np.inf
        else:
            raise ValueError(f"Mode must be 'min' or 'max': {self.mode!r} given.")

    def __repr__(self):
        class_name = self.__class__.__name__
        s = f"metric={self.metric!r}, mode={self.mode!r}"
        if self.best_epoch != -1:
            s += f", best_score={self.best_score:.3g}, best_epoch={self.best_epoch}"
        return f"{class_name}({s})"

    def on_epoch_end(self, run: Run):
        if self.metric not in run.metrics:
            msg = f"Metric {self.metric!r} not found. Available metrics: {run.metrics}"
            raise ValueError(msg)
        score = run.metrics[self.metric]
        if self.mode == "min":
            self.is_best = score < self.best_score
        else:
            self.is_best = score > self.best_score
        if self.is_best:
            self.best_score = score
            self.best_epoch = run.metrics.epoch
        self.score = score



from dataclasses import dataclass
from typing import Callable, Dict, Optional, Tuple, Union

import numpy as np

import ivory.core.collections
from ivory.core import instance

Index = Union[int, np.ndarray]
Value = Union[np.ndarray, Dict[str, np.ndarray]]


@dataclass
class Data:
    """Base class to provide data to `Dataset`. """

    def __post_init__(self):
        self.fold = None
        self.index = None
        self.input = None
        self.target = None
        self.init()

    def __repr__(self):
        cls_name = self.__class__.__name__
        if self.fold is None:
            return f"{cls_name}()"
        else:
            num_train = self.fold[self.fold != -1].shape[0]
            num_test = len(self.fold) - num_train
            return f"{cls_name}(train_size={num_train}, test_size={num_test})"

    def init(self):
        """Initializes `fold`, `index`, `input`, `target`."""

    def get_index(self, mode: str, fold: int) -> np.ndarray:
        """Returns index according to the mode and fold for `Dataset`.

        Args:
            mode: `train`, `val`, or `test`.
            fold: fold number

        Returns:
            data index array.
        """
        index = np.arange(len(self.fold))
        if mode == "train":
            return index[(self.fold != fold) & (self.fold != -1)]
        elif mode == "val":
            return index[self.fold == fold]
        else:
            return index[self.fold == -1]

    def get_input(self, index: Index) -> Value:
        """
        Args:
            index: Index
        """
        return self.input[index]

    def get_target(self, index: Index) -> Value:
        return self.target[index]

    def get(self, index: Union[int, np.ndarray]) -> Tuple[Index, Value, Value]:
        """Returns a tuple of (index, input, target) according to the index."""
        return self.index[index], self.get_input(index), self.get_target(index)


@dataclass
class Dataset:
    """Dataset class which implements `__len__()`, `__getitem__()`, `__iter__()`.

    Args:
        data: data from which to load the data.
        mode: `train`, `val`, or `test`.
        fold: fold number.
        transform: callable to transform the data.
    """

    data: Data
    mode: str
    fold: int
    transform: Optional[Callable] = None

    def __post_init__(self):
        self.index = self.data.get_index(self.mode, self.fold)
        if self.mode == "test":
            self.fold = -1
        self.init()
        if self.transform:
            self.transform = instance.get_attr(self.transform)

    def init(self):
        pass

    def __repr__(self):
        cls_name = self.__class__.__name__
        return f"{cls_name}(mode={self.mode!r}, num_samples={len(self)})"

    def __len__(self):
        return len(self.index)

    def __getitem__(self, index):
        if index == slice(None, None, None):
            index, input, *target = self.get()
        else:
            index, input, *target = self.get(index)
        if self.transform:
            input, *target = self.transform(self.mode, input, *target)
        return (index, input, *target)

    def __iter__(self):
        for index in range(len(self)):
            yield self[index]

    def get(self, index=None):
        if index is None:
            return self.data.get(self.index)
        else:
            return self.data.get(self.index[index])

    def sample(self, n: int = 0, frac: float = 0.0):
        index, input, *target = self[:]
        if frac:
            n = int(len(index) * frac)
        idx = np.random.permutation(len(index))[:n]
        return tuple([x[idx] for x in [index, input, *target]])


@dataclass
class Datasets(ivory.core.collections.Dict):
    data: Data
    dataset: Callable
    fold: int

    def __post_init__(self):
        super().__init__()
        for mode in ["train", "val", "test"]:
            self[mode] = self.dataset(self.data, mode, self.fold)



import numpy as np
import torch

import ivory.callbacks.results
from ivory.torch import utils


class Results(ivory.callbacks.results.Results):
    def reset(self):
        super().reset()
        self.indexes = []
        self.outputs = []
        self.targets = []

    def step(self, index, output, target=None):
        if torch.is_tensor(index):
            index = index.numpy()
        self.indexes.append(index)

        output = output.detach()
        if output.device.type != "cpu":
            output = utils.cpu(output)
        self.outputs.append(output.numpy())

        if target is not None:
            if torch.is_tensor(target):
                if target.device.type != "cpu":
                    target = utils.cpu(target)
                target = target.numpy()
            self.targets.append(target)

    def result_dict(self):
        index = np.concatenate(self.indexes)
        output = np.concatenate(self.outputs)
        if self.targets:
            target = np.concatenate(self.targets)
        else:
            target = None
        super().step(index, output, target)
        return super().result_dict()



from dataclasses import dataclass

from optuna.exceptions import TrialPruned
from termcolor import colored

from ivory.core.exceptions import EarlyStopped, Pruned
from ivory.core.run import Run
from ivory.core.state import State
from ivory.utils.tqdm import tqdm


@dataclass
class Trainer(State):
    epoch: int = -1
    epochs: int = 1
    global_step: int = -1
    verbose: int = 1

    def start(self, run: Run):
        """Starts a train or test loop.

        Args:
            run: a run instance.
        """
        if run.mode == "train":
            self.train(run)
        else:
            self.test(run)

    def train(self, run: Run):
        run.on_fit_begin()
        try:
            self.loop(run)
        finally:
            run.on_fit_end()

    def test(self, run: Run):
        self.test_loop(run)

    def loop(self, run: Run):
        max_epoch = self.epoch + self.epochs
        epochs = range(self.epoch + 1, max_epoch + 1)
        if self.verbose == 1:
            epochs = tqdm(epochs, desc="Epoch", leave=False)
        early_stopped = pruned = None
        for self.epoch in epochs:
            if early_stopped or pruned:  # for tqdm
                continue
            run.on_epoch_begin()
            self.train_loop(run)
            self.val_loop(run)
            try:
                run.on_epoch_end()
            except EarlyStopped as e:
                early_stopped = e
            except Pruned as e:
                pruned = e
            except TrialPruned as e:
                pruned = e
            finally:
                if self.verbose:
                    self.log(run, early_stopped, pruned)
        if isinstance(pruned, TrialPruned):
            raise pruned

    def get_dataloader(self, run: Run, mode: str):
        raise NotImplementedError

    def tqdm(self, dataloader, mode):
        if self.verbose == 1:
            mode = "%-5s" % (mode[0].upper() + mode[1:])
            dataloader = tqdm(dataloader, desc=mode, leave=False)
        return dataloader

    def train_loop(self, run: Run):
        run.on_train_begin()
        dataloader = self.tqdm(self.get_dataloader(run, "train"), "train")
        for index, input, target in dataloader:
            self.global_step += 1
            self.train_step(run, index, input, target)
        run.on_train_end()

    def val_loop(self, run: Run):
        run.on_val_begin()
        dataloader = self.tqdm(self.get_dataloader(run, "val"), "val")
        for index, input, target in dataloader:
            self.val_step(run, index, input, target)
        run.on_val_end()

    def test_loop(self, run: Run):
        run.on_test_begin()
        dataloader = self.tqdm(self.get_dataloader(run, "test"), "test")
        for index, input, *target in dataloader:
            self.test_step(run, index, input, *target)
        run.on_test_end()

    def train_step(self, run: Run, index, input, target):
        """Performs a single train step."""

    def val_step(self, run: Run, index, input, target):
        """Performs a single validation step."""

    def test_step(self, run: Run, index, input, *target):
        """Performs a single test step."""

    def log(self, run: Run, early_stopped=False, pruned=False):
        msg = message(run, early_stopped, pruned)
        if msg:
            tqdm.write(msg)


def message(run: Run, early_stopped=None, pruned=None) -> str:
    if not run.metrics:
        return ""
    msg = f"[epoch#{run.trainer.epoch}]"
    if not run.monitor:
        pass
    elif run.monitor.is_best:
        msg = colored(msg, "green")
    else:
        msg = colored(msg, "yellow")
    msg += f" {run.metrics}"
    if run.monitor and run.monitor.is_best:
        msg += colored(" best", "green")
    if early_stopped:
        msg += colored(" early stopped", "magenta")
    if pruned:
        msg += colored(" pruned", "red")
    return msg



"""Pruning class to prune unpromising trials."""
from dataclasses import dataclass
from typing import Optional

import numpy as np
import optuna
from optuna.trial import Trial

from ivory.core.exceptions import Pruned
from ivory.core.run import Run


@dataclass
class Pruning:
    """Callback to prune unpromising trials.

    Args:
        trial:
            A `Trial` corresponding to the current evaluation of the
            objective function.
        metric:
            An evaluation metric for pruning, e.g., `val_loss`
    """

    trial: Optional[Trial] = None
    metric: str = ""

    def on_epoch_end(self, run: Run):
        if self.trial is not None:
            score = run.metrics[self.metric]
            if np.isnan(score):
                return
            epoch = run.metrics.epoch
            self.trial.report(score, step=epoch)
            if self.trial.should_prune():
                message = f"Trial was pruned at epoch {epoch}."
                raise optuna.exceptions.TrialPruned(message)

        if run.tracking:
            status = run.tracking.client.get_run(run.id).info.status
            if status == "KILLED":
                raise Pruned



from dataclasses import dataclass, field
from typing import Optional

import optuna
from optuna.pruners import BasePruner
from optuna.samplers import BaseSampler

from ivory.core.run import Study

DEFAULT_STORAGE = "sqlite://"


@dataclass
class Tuner:
    storage: str = field(default_factory=lambda: DEFAULT_STORAGE)
    sampler: Optional[BaseSampler] = None
    pruner: Optional[BasePruner] = None
    load_if_exists: bool = True

    def create_study(self, study_name: str, mode: str) -> Study:
        """Creates and returns an Optuna `Study` object."""
        if mode == "min":
            direction = "minimize"
        elif mode == "max":
            direction = "maximize"
        else:
            raise ValueError("Mode must be 'min' or 'max'.")
        study = optuna.create_study(
            storage=self.storage,
            sampler=self.sampler,
            pruner=self.pruner,
            study_name=study_name,
            direction=direction,
            load_if_exists=self.load_if_exists,
        )
        return study

    def delete_study(self, study_name: str):
        optuna.delete_study(storage=self.storage, study_name=study_name)



import ivory.core.estimator
from ivory.core import instance
from ivory.core.exceptions import EarlyStopped, Pruned
from ivory.core.run import Run
from ivory.core.trainer import message
from ivory.tensorflow.callback import Callback
from ivory.utils.tqdm import tqdm


class Trainer(ivory.core.estimator.Estimator):
    def __init__(self, epoch=-1, epochs=1, batch_size=32, verbose=1, **kwargs):
        self.epoch = epoch
        self.epochs = epochs
        self.batch_size = batch_size
        self.verbose = verbose
        self.kwargs = kwargs

    def __repr__(self):
        class_name = self.__class__.__name__
        return f"{class_name}()"

    def on_init_begin(self, run):
        if not run.model._is_compiled:
            params, self.kwargs = instance.filter_params(run.model.fit, **self.kwargs)
            if run.optimizer:
                params.update(optimizer=run.optimizer)
            run.model.compile(**params)

    def train(self, run: Run):
        run.on_fit_begin()
        callbacks = [Callback(run)]
        if run.scheduler:
            callbacks.insert(0, run.scheduler)
        if "callbacks" in self.kwargs:
            callbacks = self.kwargs.pop("callbacks") + callbacks
        train_dataset = run.datasets.train[:][1:]
        val_dataset = run.datasets.val[:][1:]
        try:
            run.model.fit(
                *train_dataset,
                validation_data=val_dataset,
                batch_size=self.batch_size,
                initial_epoch=self.epoch + 1,
                epochs=self.epoch + self.epochs + 1,
                callbacks=callbacks,
                verbose=0,
                **self.kwargs,
            )
        except (EarlyStopped, Pruned):
            pass
        finally:
            run.on_fit_end()

    def test(self, run: Run):
        run.on_test_begin()
        index, input, *target = run.datasets.test[:]
        output = run.model.predict(input, callbacks=[Callback(run)])
        if run.results:
            run.results.step(index, output, *target)
        run.on_test_end()

    def log(self, run: Run, early_stopped=False, pruned=False):
        msg = message(run, early_stopped, pruned)
        if msg:
            tqdm.write(msg)

    def on_fit_begin(self, run: Run):
        if self.verbose == 1:
            self.epoch_bar = tqdm(total=self.epochs, desc="Epoch", leave=False)

    def on_epoch_end(self, run: Run):
        if self.verbose == 1:
            self.epoch_bar.update(1)

    def on_fit_end(self, run: Run):
        if self.verbose == 1:
            self.epoch_bar.close()

    def on_loop_begin(self, run: Run, mode: str):
        if self.verbose == 1:
            dataset = run.datasets[mode]
            total = len(dataset) // self.batch_size
            mode = "%-5s" % (mode[0].upper() + mode[1:])
            self.batch_bar = tqdm(total=total, desc=mode, leave=False)

    def on_batch_end(self):
        if self.verbose == 1:
            self.batch_bar.update(1)

    def on_train_begin(self, run: Run):
        self.on_loop_begin(run, "train")

    def on_train_end(self, run: Run):
        if self.verbose == 1:
            self.batch_bar.close()

    def on_val_begin(self, run: Run):
        self.on_loop_begin(run, "val")

    def on_val_end(self, run: Run):
        if self.verbose == 1:
            self.batch_bar.close()

    def on_test_begin(self, run: Run):
        self.on_loop_begin(run, "test")

    def on_test_end(self, run: Run):
        if self.verbose == 1:
            self.batch_bar.close()

    def state_dict(self):
        state_dict = super().state_dict()
        if "epoch_bar" in state_dict:
            del state_dict["epoch_bar"]
        if "batch_bar" in state_dict:
            del state_dict["batch_bar"]
        return state_dict



from typing import List

import numpy as np

import ivory.callbacks.metrics
from ivory.core.run import Run


class Metrics(ivory.callbacks.metrics.Metrics):
    def on_epoch_begin(self, run: Run):
        self.epoch = run.trainer.epoch

    def on_train_begin(self, run: Run):
        self.losses: List[float] = []

    def step(self, loss: float):
        self.losses.append(loss)

    def on_train_end(self, run: Run):
        self["loss"] = np.mean(self.losses)

    def on_val_begin(self, run: Run):
        self.losses = []

    def on_val_end(self, run: Run):
        self["val_loss"] = np.mean(self.losses)

    def metrics_dict(self, run):
        return {"lr": run.optimizer.param_groups[0]["lr"]}

    # def save(self, state_dict, path):
    #     torch.save(state_dict, path)
    #
    # def load(self, path):
    #     return torch.load(path)



import numpy as np
from sklearn.model_selection import GroupKFold, KFold, StratifiedKFold


def fold_array(splitter, x, y=None, groups=None):
    fold = np.full(x.shape[0], -1, dtype=np.int8)
    for i, (_, test_index) in enumerate(splitter.split(x, y, groups)):
        fold[test_index] = i
    return fold


def kfold_split(x, n_splits=5, shuffle=True, random_state=0):
    splitter = KFold(n_splits, shuffle=shuffle, random_state=random_state)
    return fold_array(splitter, x)


def group_kfold_split(groups, n_splits=5):
    splitter = GroupKFold(n_splits)
    return fold_array(splitter, groups, groups, groups=groups)


def stratified_kfold_split(y, n_splits=5, shuffle=True, random_state=0):
    splitter = StratifiedKFold(n_splits, shuffle=shuffle, random_state=random_state)
    return fold_array(splitter, y, y)


def multilabel_stratified_kfold_split(labels, n_splits, shuffle=True, random_state=0):
    from iterstrat.ml_stratifiers import MultilabelStratifiedKFold

    splitter = MultilabelStratifiedKFold(
        n_splits, shuffle=shuffle, random_state=random_state
    )
    x = np.arange(len(labels))
    return fold_array(splitter, x, labels)



import pickle
from typing import Any, Dict


class State:
    def state_dict(self) -> Dict[str, Any]:
        state_dict = {}
        for key, value in self.__dict__.items():
            if not callable(value):
                state_dict[key] = value
        return state_dict

    def load_state_dict(self, state_dict: Dict[str, Any]):
        self.__dict__.update(state_dict)


def save(state_dict: Dict[str, Any], path: str):
    with open(path, "wb") as file:
        pickle.dump(state_dict, file)


def load(path: str) -> Dict[str, Any]:
    with open(path, "rb") as file:
        return pickle.load(file)



import itertools
from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple


def update_dict(org: Dict[str, Any], update: Dict[str, Any]) -> None:
    """Update dict using dot-notation.

    Examples:
        >>> x = {"a": 1, "b": {"x": "abc", "y": 2, "z": [0, 1, 2]}}
        >>> update_dict(x, {"b": {"z": [0, 1]}, "b.x": "def"})
        >>> x
        {'a': 1, 'b': {'x': 'def', 'y': 2, 'z': [0, 1]}}
        >>> update_dict(x, {"b.z.1": 10, "b.x": "def"})
        >>> x
        {'a': 1, 'b': {'x': 'def', 'y': 2, 'z': [0, 10]}}
    """
    for key, value in update.items():
        x = org
        attrs = key.split(".")
        for attr in attrs[:-1]:
            x = x[attr]
        k = attrs[-1]
        if k not in x:
            if "0" <= k[0] <= "9":
                k = int(k)  # type:ignore
            x[k] = value
        elif isinstance(x[k], str) and x[k].startswith("$"):
            x[k] = value
        elif type(x[k]) is not type(value) and x[k] is not None:
            raise ValueError(f"different type: {type(x[k])} != {type(value)}")
        else:
            if isinstance(x[k], dict):
                x[k].update(value)
            else:
                x[k] = value


def colon_to_list(x: Dict[str, Any]) -> Dict[str, Any]:
    """Converts suffix integers into a list.

    Examples:
        >>> x = {"a:0": 1, "a:1": 3, "b.x:0": 10, "b.x:1": 20, "c": 100}
        >>> colon_to_list(x)
        {'a': [1, 3], 'b.x': [10, 20], 'c': 100}
    """
    update: Dict[str, Any] = {}
    for key, value in x.items():
        if ":" not in key:
            update[key] = value
            continue
        key, digit = key.split(":")
        index = int(digit)
        if index == 0:
            if key in update:
                raise KeyError(key)
            update[key] = [value]
        elif key not in update or len(update[key]) != index:
            raise KeyError(key)
        else:
            update[key].append(value)
    return update


def dot_flatten(x: Dict[str, Any], flattened=None, prefix="") -> Dict[str, Any]:
    """Flatten dict in dot-format.

    Examples:
        >>> params = {"model": {"name": "abc", "x": {"a": 1, "b": 2}}}
        >>> dot_flatten(params)
        {'model.name': 'abc', 'model.x.a': 1, 'model.x.b': 2}
    """
    if flattened is None:
        flattened = {}
    for key, value in x.items():
        if isinstance(value, dict):
            dot_flatten(x[key], flattened, prefix + key + ".")
        else:
            flattened[prefix + key] = value
    return flattened


def dot_get(x: Dict[str, Any], key: str):
    """Dot style dictionay access.

    Examples:
        >>> x = {"a": 1, "b": {"x": "abc", "y": 2, "z": [0, 1, 2]}}
        >>> dot_get(x, "a")
        1
        >>> dot_get(x, "b.x")
        'abc'
        >>> dot_get(x, "b.z.1")
        1
    """
    keys = key.split(".")
    for key in keys[:-1]:
        if key not in x:
            return None
        x = x[key]
    key = keys[-1]
    if "0" <= key[0] <= "9":
        return x[int(key)]  # type:ignore
    else:
        return x[key]


def get_fullnames(params, name, prefix="", dict_allowed=False):
    """Returns a fullname found first.

    Examples:
        >>> params = {'a': 1, 'b': {'c': {'d': 2, 'e': [1, 2, 3]}}, "x": {'d': 2}}
        >>> list(get_fullnames(params, 'a'))
        ['a']
        >>> list(get_fullnames(params, 'c'))
        []
        >>> list(get_fullnames(params, 'd'))
        ['b.c.d', 'x.d']
        >>> list(get_fullnames(params, 'e'))
        ['b.c.e']
        >>> list(get_fullnames(params, 'b.c.d'))
        ['b.c.d']
        >>> list(get_fullnames(params, 'c.d'))
        ['b.c.d']
        >>> list(get_fullnames(params, 'e.2'))
        ['b.c.e.2']
    """
    if "." in name:
        name, _, suffix = name.partition(".")
        for fullname in get_fullnames(params, name, dict_allowed=True):
            yield ".".join([fullname, suffix])
    elif not isinstance(params, dict):
        return
    elif name in params:
        if dict_allowed or not isinstance(params[name], dict):
            yield prefix + name
    else:
        for key in params:
            prefix_ = prefix + key + "."
            yield from get_fullnames(params[key], name, prefix_, dict_allowed)


def create_update(params, args: Optional[Dict[str, Any]] = None, **kwargs):
    if args is None:
        args = {}
    args.update(kwargs)
    args = colon_to_list(args)
    args_ = {}
    update = {}
    for name, value in args.items():
        for fullname in get_fullnames(params, name):
            update[fullname] = value
            args_[name] = value
    return update, args_


def get_value(params, name):
    """
    Examples:
        >>> params = {'a': 1, 'b': {'c': {'d': 2}}}
        >>> get_value(params, 'a')
        1
        >>> get_value(params, 'd')
        2
        >>> get_value(params, 'b.c.d')
        2
        >>> get_value(params, 'c.d')
        2
    """
    fullnames = list(get_fullnames(params, name))
    if fullnames:
        return dot_get(params, fullnames[0])


def match(params, **query):
    """Returns if params match the query or not.

    Avaliable query type:
        tuple: (start, top) range including the stop value.
        list: [a1, a2, ..., an] parameters set.
        other: exact match.

    Examples:
        >>> params = {'a': 1, 'b': {'c': {'d': 2}}}
        >>> match(params, a=1)
        True
        >>> match(params, a=2)
        False
        >>> match(params, a=(0, 3))
        True
        >>> match(params, a=(3, 4))
        False
        >>> match(params, a=[5, 6])
        False
        >>> match(params, a=[0, 1])
        True
    """
    for name, cond in query.items():
        value = get_value(params, name)
        if value is None:
            return False
        elif isinstance(cond, tuple):
            if value < cond[0] or value > cond[1]:
                return False
        elif isinstance(cond, list):
            if value not in cond:
                return False
        elif value != cond:
            return False
    return True


def product(params: Dict[str, Iterable[Any]]) -> Iterator[Dict[str, Any]]:
    """
    Examples:
        >>> params = {"a": [1, 2], "b": [3, 4]}
        >>> list(product(params))
        [{'a': 1, 'b': 3}, {'a': 1, 'b': 4}, {'a': 2, 'b': 3}, {'a': 2, 'b': 4}]
    """
    for values in itertools.product(*params.values()):
        args = {}
        for name, value in zip(params.keys(), values):
            args[name] = value
        yield args


def to_str(params):
    t = []
    for key, value in params.items():
        if key == "verbose":
            continue
        if isinstance(value, str):
            t.append(f"{key}={value!r}")
        else:
            t.append(f"{key}={value:.4g}")
    return " ".join(t)


def split_params(
    params: Optional[Dict[str, Iterable[Any]]] = None, **kwargs
) -> Tuple[Dict[str, Iterable[Any]], Dict[str, Any]]:
    """Returns (`params`, `base_params`).

    `params` for variable parameters and `base_params` for fixed parameters.

    Args:
        params: Parameter dictionay.
        **kwargs: Optional parameters

    Examples:
        >>> params = {"a": range(2), "b": 3}
        >>> params, base_params = split_params(params, c='abc', d=range(3))
        >>> params
        {'a': range(0, 2), 'd': range(0, 3)}
        >>> base_params
        {'b': 3, 'c': 'abc'}
    """
    if params is None:
        params = {}
    if kwargs:
        params.update(kwargs)
    params = params.copy()
    base_params: Dict[str, Any] = {}
    params_list = {arg: to_list(values) for arg, values in params.items()}
    for arg, values in params_list.items():
        if len(values) == 1:
            base_params[arg] = values[0]
            del params[arg]
    return params, base_params


def to_list(x) -> List[Any]:
    """
    Examples:
        >>> to_list(range(3))
        [0, 1, 2]
        >>> to_list(100)
        [100]
        >>> to_list('abc')
        ['abc']
    """
    if isinstance(x, str):
        return [x]
    try:
        return list(x)
    except TypeError:
        return [x]



from tensorflow.python.keras import backend as K

import ivory.callbacks.metrics
from ivory.core.run import Run


class Metrics(ivory.callbacks.metrics.Metrics):
    def metrics_dict(self, run: Run):
        return {"lr": float(K.get_value(run.model.optimizer.lr))}



"""Early stopping when a monitored metric has stopped improving."""
from dataclasses import dataclass

from ivory.core.exceptions import EarlyStopped
from ivory.core.run import Run
from ivory.core.state import State


@dataclass
class EarlyStopping(State):
    """Early stops a training loop when a monitored metric has stopped improving.

    Args:
        patience: number of epochs with no improvement after which training will be
            stopped.

    Raises:
        EarlyStopped: when ealry stopping occurs.
    """

    patience: int
    wait: int = 0

    # def __post_init__(self):
    #     self.wait = 0

    def on_epoch_end(self, run: Run):
        if run.monitor.is_best:
            self.wait = 0
        else:
            self.wait += 1
            if self.wait >= self.patience:
                raise EarlyStopped



import ivory.core.estimator
from ivory.core import instance


class Estimator(ivory.core.estimator.Estimator):
    def __init__(self, model, return_probability=True, **kwargs):
        model = instance.get_attr(model)
        params, kwargs = instance.filter_params(model, **kwargs)
        if params:
            raise ValueError(f"Unknown parameters: {list(params.keys())}")
        self.estimator = model(**kwargs)
        if not hasattr(self.estimator, "predict_proba"):
            return_probability = False
        self.return_probability = return_probability

    def predict(self, input):
        if self.return_probability:
            return self.estimator.predict_proba(input)
        else:
            return self.estimator.predict(input)



import tensorflow.keras.callbacks
from optuna.exceptions import TrialPruned

from ivory.core.exceptions import EarlyStopped, Pruned
from ivory.core.run import Run


class Callback(tensorflow.keras.callbacks.Callback):
    """Callback class used to connect Ivory and Tensorflow's original `Callback`.

    Args:
        run: A `Run` instance connected to this `Callback`.
    """

    def __init__(self, run: Run):
        self.run = run
        self.trainer = run.trainer
        self.predict = run.model.predict

    def on_epoch_begin(self, epoch, logs=None):
        """Calls `run.on_epoch_begin()` and `run.on_train_begin()`."""
        self.trainer.epoch = epoch
        self.run.on_epoch_begin()
        self.run.on_train_begin()

    def on_test_begin(self, logs=None):
        """Calls `run.on_train_end()` and `run.on_val_begin()`."""
        self.trainer.step(self.run, "train", training=False, predict=self.predict)
        self.run.on_train_end()
        self.run.on_val_begin()

    def on_test_end(self, logs=None):
        """Calls `run.on_val_end()`."""
        self.trainer.step(self.run, "val", predict=self.predict)
        self.run.on_val_end()

    def on_epoch_end(self, epoch, logs=None):
        """Calls `run.on_epoch_end()`."""
        if self.run.metrics and logs:
            for key, value in logs.items():
                key = key.replace("accuracy", "acc")
                self.run.metrics[key] = value
        early_stopped = pruned = None
        try:
            self.run.on_epoch_end()
        except EarlyStopped as e:
            early_stopped = e
        except Pruned as e:
            pruned = e
        except TrialPruned as e:
            pruned = e
        finally:
            self.trainer.log(self.run, early_stopped, pruned)
        if pruned:
            raise pruned

    def on_train_batch_end(self, batch, logs=None):
        """Call `trainer.on_batch_end` to update a progress bar."""
        self.trainer.on_batch_end()

    def on_test_batch_end(self, batch, logs=None):
        """Call `trainer.on_batch_end` to update a progress bar."""
        self.trainer.on_batch_end()

    def on_predict_batch_end(self, batch, logs=None):
        """Call `trainer.on_batch_end` to update a progress bar."""
        self.trainer.on_batch_end()



import sklearn.metrics

import ivory.callbacks.metrics


class Metrics(ivory.callbacks.metrics.Metrics):
    def metrics_dict(self, run):
        pred = run.results.val.output.reshape(-1)
        true = run.results.val.target.reshape(-1)
        estimator = run.estimator.estimator
        metrics_dict = {}
        if hasattr(estimator, "criterion"):
            if estimator.criterion == "mse":
                metrics_dict["mse"] = sklearn.metrics.mean_squared_error(true, pred)
            elif estimator.criterion == "mae":
                metrics_dict["mae"] = sklearn.metrics.mean_absolute_error(true, pred)
        metrics_dict.update(super().metrics_dict(run))
        return metrics_dict



import ast
import re

import numpy as np


class Range:
    def __init__(self, start, stop=None, step=1, num: int = 0, log: bool = False):
        if isinstance(start, str):
            start, stop, step, num, log = parse(start)
        elif stop is None:
            start, stop = 0, start
        self.start = start
        self.stop = stop
        self.step = step
        self.num = num
        self.log = log
        if log and step != 1:
            raise ValueError("Invalid step.")

    @property
    def is_integer(self):
        return all(isinstance(x, int) for x in [self.start, self.stop, self.step])

    @property
    def is_float(self):
        return not self.is_integer

    def __repr__(self):
        class_name = self.__class__.__name__
        s = f"{class_name}({self.start}, {self.stop}"
        if self.step != 1:
            s += f", {self.step}"
        if self.num >= 2:
            s += f", n={self.num}"
        if self.log:
            s += f", log={self.log}"
        return s + ")"

    def __iter__(self):
        if self.is_integer:
            if self.start < self.stop:
                it = range(self.start, self.stop + 1, self.step)
            else:
                it = range(self.start, self.stop - 1, -self.step)
            if self.num < 2:
                return iter(it)
            else:
                values = list(it)
                index = np.linspace(0, len(values) - 1, self.num)
                return (values[int(round(x))] for x in index)
        else:
            num = self.num
            if self.log:
                if self.num < 2:
                    raise ValueError(f"num must be larger than 1, but {num} given.")
                start = np.log10(self.start)
                stop = np.log10(self.stop)
                return iter(float(x) for x in np.logspace(start, stop, num))
            else:
                if num < 2:
                    num = round(abs(self.stop - self.start) / self.step + 1)
                return iter(float(x) for x in np.linspace(self.start, self.stop, num))

    def __len__(self):
        return len(list(iter(self)))


def parse(value: str):
    """
    Examples:
        >>> parse('2-3')
        (2, 3, 1, 0, False)
        >>> parse('2-4-2')
        (2, 4, 2, 0, False)
        >>> parse('2-4:3')
        (2, 4, 1, 3, False)
        >>> parse('0-1-0.2')
        (0, 1, 0.2, 0, False)
        >>> parse('0-1:5')
        (0, 1, 1, 5, False)
        >>> parse('0-1.log')
        (0, 1, 1, 0, True)
        >>> parse('1e-3_1e-2.log')
        (0.001, 0.01, 1, 0, True)
    """
    num = 0
    log = False
    if value.endswith(".log"):
        value = value[:-4]
        log = True
    if ":" in value:
        value, num_str = value.split(":")
        num = int(num_str)
    sep = "_" if "_" in value else "-"
    match = re.match(f"(.+){sep}(.+)", value)
    if not match:
        raise ValueError(f"Invalid string for Range: {value}")
    if sep in match.group(1):
        if num != 0:
            raise ValueError(f"Invalid string for Range: {value}")
        start, stop = match.group(1).split(sep)
        step = match.group(2)
    else:
        start = match.group(1)
        stop = match.group(2)
        step = "1"
    start = literal_eval(start)
    stop = literal_eval(stop)
    step = literal_eval(step)
    if all(isinstance(x, (int, float)) for x in [start, stop, step]):
        return start, stop, step, num, log
    raise ValueError(f"Invalid string for Range: {value}")


def literal_eval(value):
    try:
        return ast.literal_eval(value)
    except ValueError:
        return value



from ivory.core.run import Run
from ivory.core.state import State
from ivory.utils.tqdm import tqdm


class Estimator(State):
    def start(self, run: Run):
        if run.mode == "train":
            self.train(run)
        else:
            self.test(run)

    def train(self, run: Run):
        run.on_fit_begin()
        run.on_epoch_begin()
        run.on_train_begin()
        self.step(run, "train")
        run.on_train_end()
        run.on_val_begin()
        self.step(run, "val")
        run.on_val_end()
        run.on_epoch_end()
        self.log(run)
        run.on_fit_end()

    def test(self, run: Run):
        run.on_test_begin()
        self.step(run, "test")
        run.on_test_end()

    def step(self, run: Run, mode: str, training: bool = True, predict=None):
        index, input, *target = run.datasets[mode][:]
        if mode == "train" and training:
            self.fit(input, *target)
        predict = predict or self.predict
        output = predict(input)
        if run.results:
            run.results.step(index, output, *target)

    def fit(self, input, target):
        self.estimator.fit(input, target)

    def predict(self, input):
        return self.estimator.predict(input)

    def log(self, run: Run):
        if run.metrics:
            metrics = str(run.metrics)
            if metrics:
                tqdm.write(f"[{run.name}] {metrics}")



import os

import tensorflow as tf

import ivory.core.run


class Run(ivory.core.run.Run):
    def save(self, directory: str):
        super().save(directory)
        if self.model:
            path = os.path.join(directory, "model/checkpoint")
            level = tf.get_logger().level
            tf.get_logger().setLevel("WARNING")
            self.model.save_weights(path)
            tf.get_logger().setLevel(level)

    def load_instance(self, path):
        path = os.path.join(path, "checkpoint")
        if self.model:
            self.model.load_weights(path)



import logging
import sys

import click
import logzero
from logzero import logger

import ivory
from ivory.core import parser

# from ivory.utils.range import Range

if "." not in sys.path:
    sys.path.insert(0, ".")


def loglevel(ctx, param, value):
    if param.name == "quiet" and value is True:
        logzero.loglevel(logging.WARNING)
    elif param.name == "verbose" and value is True:
        logzero.loglevel(logging.DEBUG)
    else:
        logzero.loglevel(logging.INFO)
    return value


@click.group()
def cli():
    pass


@cli.command(help="Invoke a run or product runs.")
@click.argument("name")
@click.argument("args", nargs=-1)
@click.option("-r", "--repeat", default=1, help="Number of repeatation.")
@click.option("-n", "--number", default=None, help="Task number to load.")
@click.option("-q", "--quiet", is_flag=True, help="Queit mode.", callback=loglevel)
@click.option("-v", "--verbose", is_flag=True, help="Verbose mode.", callback=loglevel)
def run(name, args, repeat, number, quiet, verbose):
    client = ivory.create_client()
    if not args and repeat == 1:
        run = client.create_run(name)
        run.start("both")
    else:
        task = client.create_task(name, number and int(number))
        params = parser.parse_args(args)
        for run in task.product(params, repeat=repeat):
            run.start("both")


@cli.command(help="Optimize hyper parameters.")
@click.argument("name")
@click.argument("args", nargs=-1)
@click.option("-n", "--number", default=None, help="Study number to load.")
@click.option("-q", "--quiet", is_flag=True, help="Queit mode.", callback=loglevel)
@click.option("-v", "--verbose", is_flag=True, help="Verbose mode.", callback=loglevel)
def optimize(name, args, number, quiet, verbose):
    client = ivory.create_client()
    run_number = number and int(number)
    if "=" not in args[0]:
        study = client.create_study(name, run_number=run_number)
        suggest_name = args[0]
        params = parser.parse_args(args[1:])
        kwargs = {key: values[0] for key, values in params.items()}
        if "n_trials" not in kwargs and "timeout" not in kwargs:
            kwargs["n_trials"] = 3  # FIXME
        study.optimize(suggest_name, **kwargs)
    else:
        params = parser.parse_args(args)  # FIXME: exclude len(params)==1
        params = list(zip(*params.items()))
        params = {params[0]: params[1]}
        study = client.create_study(name, params, run_number=run_number)
        study.optimize(n_trials=3)  # FIXME


@cli.command(help="Kill the last run to prune.")
@click.argument("name")
def kill(name):
    client = ivory.create_client()
    client.set_terminated(name, status="KILLED", run=-1)


@cli.command(help="Update parameters.")
@click.argument("name")
def update(name):
    client = ivory.create_client()
    client.update_params(name)


@cli.command(help="Remove deleted runs.")
@click.argument("name")
def clean(name):
    client = ivory.create_client()
    num_runs = client.remove_deleted_runs(name)
    print("Removed runs:", num_runs)


@cli.command(help="Start tracking UI.")
@click.option("-q", "--quiet", is_flag=True, help="Queit mode.", callback=loglevel)
@click.option("-v", "--verbose", is_flag=True, help="Verbose mode.", callback=loglevel)
def ui(quiet, verbose):
    logger.info("Tracking UI.")
    client = ivory.create_client()
    client.ui()


def main():
    cli()


if __name__ == "__main__":
    main()



import os
import tempfile
import time
from dataclasses import dataclass
from typing import Any, Dict

import mlflow
import yaml
from mlflow.entities import Metric, Param
from mlflow.utils.mlflow_tags import MLFLOW_PARENT_RUN_ID

from ivory import utils
from ivory.core.run import Run


@dataclass
class Tracking:
    tracking_uri: str

    def __post_init__(self):
        self.client = mlflow.tracking.MlflowClient(self.tracking_uri)

    def on_epoch_end(self, run: Run):
        self.save_run(run, "current")
        if not run.metrics:
            return
        metrics = run.metrics.copy()
        monitor = run.monitor
        if monitor:
            metrics.update(best_score=monitor.best_score, best_epoch=monitor.best_epoch)
        self.log_metrics(run.id, metrics, run.metrics.epoch)

    def on_fit_end(self, run: Run):
        if self.client.get_run(run.id).info.status == "RUNNING":
            self.client.set_terminated(run.id)

    def on_test_end(self, run: Run):
        self.save_run(run, "test")

    def log_params_artifact(self, run: Run):
        with tempfile.TemporaryDirectory() as tmpdir:
            path = os.path.join(tmpdir, "params.yaml")
            with open(path, "w") as file:
                yaml.dump(run.params, file, sort_keys=False)
            with utils.path.chdir(run.source_name):
                self.client.log_artifacts(run.id, tmpdir)

    def log_params(self, run_id: str, params: Dict[str, Any]):
        params_list = []
        for key, value in params.items():
            if key != 'verbose':
                params_list.append(Param(key, to_str(value)))
        self.client.log_batch(run_id, metrics=[], params=params_list, tags=[])

    def log_metrics(self, run_id: str, metrics: Dict[str, float], step: int = 0):
        ts = int(time.time() * 1000)  # timestamp in milliseconds.
        metrics_ = [Metric(key, value, ts, step) for key, value in metrics.items()]
        self.client.log_batch(run_id, metrics=metrics_, params=[], tags=[])

    def save_run(self, run: Run, mode: str):
        with tempfile.TemporaryDirectory() as tmpdir:
            directory = os.path.join(tmpdir, mode)
            os.mkdir(directory)
            run.save(directory)
            with utils.path.chdir(run.source_name):
                self.client.log_artifacts(run.id, tmpdir)
                if mode != "current":
                    return
                if run.monitor and run.monitor.is_best and run.monitor.best_epoch > -1:
                    os.rename(directory, directory.replace("current", "best"))
                    self.client.log_artifacts(run.id, tmpdir)

    def set_tags(self, run_id: str, tags: Dict[str, Any]):
        for key, value in tags.items():
            if key != 'verbose':
                self.client.set_tag(run_id, key, to_str(value))

    def set_parent_run_id(self, run_id: str, parent_run_id: str):
        self.client.set_tag(run_id, MLFLOW_PARENT_RUN_ID, parent_run_id)


def to_str(value: Any) -> str:
    if isinstance(value, (list, tuple)):
        return "[" + ", ".join(to_str(x) for x in value) + "]"
    elif isinstance(value, float):
        return f"{value:.4g}"
    else:
        return str(value)



"""The base module contains base classes for the Ivory system."""
import copy
import inspect
from typing import Callable, Dict

import ivory.core.collections
from ivory import utils
from ivory.core import default, instance


class Base(ivory.core.collections.Dict):
    """Base class for entities such as `client`, `experiment`, and `run`. """

    def __init__(self, params=None, **objects):
        super().__init__()
        self.params = params
        self.id = self.name = self.source_name = ""
        if "id" in objects:
            self.id = objects.pop("id")
        if "name" in objects:
            self.name = objects.pop("name")
        if "source_name" in objects:
            self.source_name = objects.pop("source_name")
        self.dict = objects

    def __repr__(self):
        args = []
        if self.id:
            args.append(f"id={self.id!r}")
        if self.name:
            args.append(f"name={self.name!r}")
        args.append(f"num_objects={len(self)}")
        args = ", ".join(args)
        return f"{self.__class__.__name__}({args})"


class Creator(Base):
    """Creator class to create `run` objects."""

    @property
    def experiment_id(self):
        return self.params["experiment"]["id"]

    @property
    def experiment_name(self):
        return self.params["experiment"]["name"]

    def create_params(self, args=None, name: str = "run", **kwargs):
        params = copy.deepcopy(self.params)
        if name not in params:
            params.update(default.get(name))
        update, args = utils.params.create_update(params[name], args, **kwargs)
        utils.params.update_dict(params[name], update)
        return params, args

    def create_run(self, args=None, name: str = "run", **kwargs):
        params, args = self.create_params(args, name, **kwargs)
        run = instance.create_base_instance(params, name, self.source_name)
        if self.tracker:
            from ivory.callbacks.pruning import Pruning

            run.set_tracker(self.tracker, name)
            run.tracking.log_params_artifact(run)
            args = {arg: utils.params.get_value(run.params[name], arg) for arg in args}
            run.tracking.log_params(run.id, args)
            run.set(pruning=Pruning())
        return run

    def create_instance(self, instance_name: str, args=None, name="run", **kwargs):
        params, _ = self.create_params(args, name, **kwargs)
        return instance.create_instance(params[name], instance_name)


class Callback:
    """Callback class for the Ivory callback system."""

    METHODS = [
        "on_init_begin",
        "on_init_end",
        "on_fit_begin",
        "on_epoch_begin",
        "on_train_begin",
        "on_train_end",
        "on_val_begin",
        "on_val_end",
        "on_epoch_end",
        "on_fit_end",
        "on_test_begin",
        "on_test_end",
    ]

    ARGUMENTS = ["run"]

    def __init__(self, caller: "CallbackCaller", methods: Dict[str, Callable]):
        self.caller = caller
        self.methods = methods

    def __repr__(self):
        class_name = self.__class__.__name__
        callbacks = list(self.methods.keys())
        return f"{class_name}({callbacks})"

    def __call__(self):
        caller = self.caller
        for method in self.methods.values():
            method(caller)


class CallbackCaller(Creator):
    """Callback caller class."""

    def create_callbacks(self):
        """Creates callback functions and store them in the dict-object."""
        for method in Callback.METHODS:
            methods = {}
            for key in self:
                if hasattr(self[key], method):
                    callback = getattr(self[key], method)
                    if callable(callback):
                        parameters = inspect.signature(callback).parameters
                        if list(parameters.keys()) == Callback.ARGUMENTS:
                            methods[key] = callback

            self[method] = Callback(self, methods)


class Experiment(Creator):
    def set_tracker(self, tracker):
        if not self.id:
            self.id = tracker.create_experiment(self.name)
            self.params["experiment"]["id"] = self.id
        self.set(tracker=tracker)

    def create_task(self):
        return self.create_run(name="task")

    def create_study(self, args=None, **suggests):
        study = self.create_run(name="study")
        if isinstance(args, str) and args in study.objective:
            study.objective.suggests = {args: study.objective.suggests[args]}
            return study
        if args or suggests:
            study.objective.update(args, **suggests)
        return study



"""A container to store training, validation and test results. """
from typing import Callable, Dict, Iterable, List, Optional

import numpy as np
import pandas as pd

import ivory.core.collections
from ivory.core.run import Run
from ivory.core.state import State


class Results(ivory.core.collections.Dict, State):
    def reset(self):
        self.index = None
        self.output = None
        self.target = None

    def on_train_begin(self, run: Run):
        self.reset()

    def on_test_begin(self, run: Run):
        self.reset()

    def step(self, index, output, target=None):
        self.index = index
        self.output = output
        self.target = target

    def on_train_end(self, run: Run):
        self["train"] = self.result_dict()
        self.reset()

    def on_val_end(self, run: Run):
        self["val"] = self.result_dict()
        self.reset()

    def on_test_end(self, run: Run):
        self["test"] = self.result_dict()
        self.reset()

    def result_dict(self):
        dict = ivory.core.collections.Dict()
        return dict(index=self.index, output=self.output, target=self.target)

    def set(self, **kwargs):
        results = {}
        for key, value in kwargs.items():
            dict = ivory.core.collections.Dict()
            if len(value) == 3:
                dict(index=value[0], output=value[1], target=value[2])
            else:
                dict(index=value[0], output=value[1], target=None)
            results[key] = dict
        super().set(**results)

    def mean(self):
        results = Results()
        for mode, result in self.items():
            index = result.index
            kwargs = {}
            for key, value in list(result.items())[1:]:
                if value.ndim == 1:
                    series = pd.Series(value, index=index)
                    value = series.groupby(level=0).mean()
                else:
                    df = pd.DataFrame(value)
                    df["index"] = index
                    value = df.groupby("index").mean()
                value.sort_index(inplace=True)
                kwargs[key] = value.to_numpy()
                kwargs["index"] = value.index.to_numpy()
            dict = ivory.core.collections.Dict()
            results[mode] = dict(**kwargs)
        return results


def stack(x: List[np.ndarray]) -> np.ndarray:
    if x[0].ndim == 1:
        return np.hstack(x)
    else:
        return np.vstack(x)


def concatenate(
    iterable: Iterable[Results],
    callback: Optional[Callable] = None,
    modes: Iterable[str] = ("val", "test"),
    reduction: str = "none",
) -> Results:
    modes = list(modes)
    indexes: Dict[str, list] = {mode: [] for mode in modes}
    outputs: Dict[str, list] = {mode: [] for mode in modes}
    targets: Dict[str, list] = {mode: [] for mode in modes}
    for results in iterable:
        for mode in modes:
            if mode not in results:
                continue
            result = results[mode]
            index, output, target = result["index"], result["output"], result["target"]
            if callback:
                index, output, target = callback(index, output, target)
            indexes[mode].append(index)
            outputs[mode].append(output)
            targets[mode].append(target)
    results = Results()
    for mode in modes:
        index = np.concatenate(indexes[mode])
        output = np.concatenate(outputs[mode])
        target = np.concatenate(targets[mode])
        dict = ivory.core.collections.Dict()
        results[mode] = dict(index=index, output=output, target=target)
    if reduction != "none":
        results = getattr(results, reduction)()
    return results



import pandas as pd
import scipy.special


def softmax(df):
    prob = scipy.special.softmax(df.to_numpy(), axis=1)
    return pd.DataFrame(prob, index=df.index)


def mean(df):
    is_series = isinstance(df, pd.Series)
    df = df.reset_index().groupby("index").mean()
    df.sort_index(inplace=True)
    df.index.name = None
    if is_series:
        df = df[0]
    return df


def argmax(df):
    pred = df.to_numpy().argmax(axis=1)
    return pd.Series(pred, index=df.index)



"""
This module provides the Ivory Client class which is one of the main classes of
Ivory library.
"""
import os
import re
import subprocess
import sys
from typing import Any, Dict, Iterable, Iterator, Optional, Tuple, Union

import ivory.callbacks.results
from ivory import utils
from ivory.callbacks.results import Results
from ivory.core import default, instance
from ivory.core.base import Base, Experiment
from ivory.core.run import Run, Study, Task
from ivory.utils.tqdm import tqdm


class Client(Base):
    """The Ivory client class.

    Attributes:
        tracker (Tracker, optional): A Tracker instance for tracking run process.
        tuner (Tuner, optional): A Tuner instance for hyperparameter tuning.
    """

    def __init__(self, params=None, **objects):
        super().__init__(params, **objects)
        self.experiments: Dict[str, Experiment] = {}

    def create_experiment(self, name: str, *args, **kwargs) -> Experiment:
        """Creates an `Experiment` according to the YAML file specified by `name`.

        Args:
            name: Experiment name.
            *args: Additional parameters.
            **kwargs: Additional parameters.
        """
        if name in self.experiments and not args and not kwargs:
            return self.experiments[name]

        params, source_name = utils.path.load_params(name, self.source_name)
        if "run" not in params:
            params = {"run": params}
        if "experiment" not in params:
            params.update(default.get("experiment"))
        if "name" not in params["experiment"]:
            params["experiment"]["name"] = name
        for value in args:
            option, _ = utils.path.load_params(value, self.source_name)
            params.update(option)
        for key, value in kwargs.items():
            option, _ = utils.path.load_params(value, self.source_name)
            if key not in option:
                option = {key: option}
            params.update(option)
        experiment = instance.create_base_instance(params, "experiment", source_name)
        if self.tracker:
            experiment.set_tracker(self.tracker)
        self.experiments[name] = experiment
        return experiment

    def create_run(self, name: str, args=None, **kwargs) -> Run:
        """Creates a `Run`.

        Args:
            name: Experiment name.
            args (dict, optional): Parameter dictionary to update the default values
                of `Experiment`.
            **kwargs: Additional parameters.
        """
        return self.create_experiment(name).create_run(args, **kwargs)

    def create_task(self, name: str, run_number: Optional[int] = None) -> Task:
        """Creates a `Task`.

        Args:
            name: Experiment name.
            run_number: If specified, load an existing task instead of creating a new
                one.
        """
        if run_number is None:
            return self.create_experiment(name).create_task()
        else:
            return self.load_run_by_name(name, task=run_number)  # type:ignore

    def create_study(
        self, name: str, args=None, run_number: Optional[int] = None, **suggests
    ) -> Study:
        """Creates a `Study`.

        Args:
            name: Experiment name.
            run_number: If specified, load an existing study instead of creating a new
                one.
        """
        if run_number is None:
            study = self.create_experiment(name).create_study(args, **suggests)
        else:
            study = self.load_run_by_name(name, study=run_number)
        if self.tuner and "storage" not in study.params["study"]["tuner"]:
            study.set(tuner=self.tuner)
        return study

    def get_run_id(self, name: str, **kwargs) -> str:
        """Returns a RunID.

        Args:
            name: Experiment name.
            **kwargs: Run name. If you want to get a run with name of 'run#5',
                **kwargs shoud be `run=5`.
        """
        run_name = list(kwargs)[0]
        run_number = kwargs[run_name]
        if run_number == -1:
            return next(self.search_run_ids(name, run_name))
        else:
            experiment_id = self.tracker.get_experiment_id(name)
            return self.tracker.get_run_id(experiment_id, run_name, run_number)

    def get_run_ids(self, name: str, **kwargs) -> Iterator[str]:
        for run_name, run_numbers in kwargs.items():
            if isinstance(run_numbers, int):
                run_numbers = [run_numbers]
            for run_number in run_numbers:
                yield self.get_run_id(name, **{run_name: run_number})

    def get_parent_run_id(self, name: str, **kwargs) -> str:
        run_id = self.get_run_id(name, **kwargs)
        return self.tracker.get_parent_run_id(run_id)

    def get_nested_run_ids(self, name: str, **kwargs) -> Iterator[str]:
        run_name = list(kwargs)[0]
        run_numbers = kwargs.pop(run_name)
        parent_run_ids = self.get_run_ids(name, **{run_name: run_numbers})
        yield from self.search_run_ids(name, parent_run_id=parent_run_ids, **kwargs)

    def set_parent_run_id(self, name: str, **kwargs):
        parent = {name: number for name, number in kwargs.items() if name != "run"}
        parent_run_id = self.get_run_id(name, **parent)
        for run_id in self.get_run_ids(name, run=kwargs["run"]):
            self.tracker.set_parent_run_id(run_id, parent_run_id)

    def get_run_name(self, run_id: str) -> str:
        return self.tracker.get_run_name(run_id)

    def get_run_name_tuple(self, run_id: str) -> Tuple[str, int]:
        return self.tracker.get_run_name_tuple(run_id)

    def search_run_ids(
        self,
        name: str = "",
        run_name: str = "",
        parent_run_id: Union[str, Iterable[str]] = "",
        parent_only: bool = False,
        nested_only: bool = False,
        exclude_parent: bool = False,
        best_score_limit: Optional[float] = None,
        **query,
    ) -> Iterator[str]:
        """Yields matching run id.

        Args:
            name: Experiment name pattern for filtering.
            run_name: Run name pattern for filtering.
            parent_run_id: If specified, search from runs which have the parent id.
            parent_only: If True, search from parent runs.
            nested_only: If True, search from nested runs.
            exclude_parent: If True, skip parent runs.
            best_score_limit: Yields runs with the best score better than this value.
            **query: Key-value pairs for filtering.

        Yields:
            run_id
        """
        for experiment in self.tracker.list_experiments():
            if name and not re.match(name, experiment.name):
                continue
            yield from self.tracker.search_run_ids(
                experiment.experiment_id,
                run_name,
                parent_run_id,
                parent_only,
                nested_only,
                exclude_parent,
                best_score_limit,
                **query,
            )

    def search_parent_run_ids(self, name: str = "", **query) -> Iterator[str]:
        yield from self.search_run_ids(name, parent_only=True, **query)

    def search_nested_run_ids(self, name: str = "", **query) -> Iterator[str]:
        yield from self.search_run_ids(name, nested_only=True, **query)

    def set_terminated(self, name: str, status: Optional[str] = None, **kwargs):
        for run_id in self.get_run_ids(name, **kwargs):
            self.tracker.client.set_terminated(run_id, status=status)

    def set_terminated_all(self, name: str = ""):
        for run_id in self.search_run_ids(name):
            self.tracker.client.set_terminated(run_id)

    def load_params(self, run_id: str) -> Dict[str, Any]:
        return self.tracker.load_params(run_id)

    def load_run(self, run_id: str, mode: str = "test") -> Run:
        return self.tracker.load_run(run_id, mode)

    def load_run_by_name(self, name: str, mode: str = "test", **kwargs) -> Run:
        run_id = self.get_run_id(name, **kwargs)
        return self.load_run(run_id, mode)

    def load_instance(self, run_id: str, instance_name: str, mode: str = "test") -> Any:
        return self.tracker.load_instance(run_id, instance_name, mode)

    def load_results(
        self,
        run_ids: Union[str, Iterable[str]],
        callback=None,
        reduction: str = "none",
        verbose: bool = True,
    ) -> Results:
        """Loads results from multiple runs and concatenates them.

        Args:
            run_ids: Multiple run ids to load.
            callback (callable): Callback function for each run. This function must take
                a `(index, output, target)` and return the same signature.
            verbose: If `True`, tqdm progress bar is displayed.

        Returns:
            A concatenated results instance.
        """
        if isinstance(run_ids, str):
            return self.load_instance(run_ids, "results")
        run_ids = list(run_ids)
        it = (self.load_instance(run_id, "results") for run_id in run_ids)
        if verbose:
            it = tqdm(it, total=len(run_ids), leave=False)
        return ivory.callbacks.results.concatenate(
            it, callback=callback, reduction=reduction
        )

    def ui(self):
        tracking_uri = self.tracker.tracking_uri
        try:
            subprocess.run(["mlflow", "ui", "--backend-store-uri", tracking_uri])
        except KeyboardInterrupt:
            pass

    def update_params(self, name: str = "", **default):
        for experiment in self.tracker.list_experiments():
            if name and not re.match(name, experiment.name):
                continue
            self.tracker.update_params(experiment.experiment_id, **default)

    def remove_deleted_runs(self, name: str = "") -> int:
        """Removes deleted runs from a local file system.

        Args:
            name: A regex pattern of experiment name for filtering.

        Returns:
            Number of removed runs.
        """
        num_runs = 0
        for experiment in self.tracker.list_experiments():
            if name and not re.match(name, experiment.name):
                continue
            num_runs += self.tracker.remove_deleted_runs(experiment.experiment_id)
        return num_runs


def create_client(
    directory: str = "", name: str = "client", tracker: bool = True
) -> Client:
    """Creates an Ivory client.

    Args:
        directory: A working directory. If a YAML file specified by the `name`
            parameter exists, the file is loaded to configure the client. In addition,
            this directory is automatically inserted to `sys.path`.
        name: A YAML config file name.
        tracker: If true, the client instance has a tracker.

    Returns:
        An created client.

    Note:
        If `tracker` is True (default value), a `mlruns` directory is made under the
        working directory by the MLFlow Tracking.
    """
    if directory:
        path = os.path.abspath(directory)
        if path not in sys.path:
            sys.path.insert(0, path)
    source_name = utils.path.normpath(name, directory)
    if os.path.exists(source_name):
        params, _ = utils.path.load_params(source_name)
    else:
        params = default.get("client")
    if not tracker and "tracker" in params["client"]:
        params["client"].pop("tracker")
    with utils.path.chdir(source_name):
        client = instance.create_base_instance(params, "client", source_name)
    return client



import copy
from typing import Any, Dict

DEFAULTS: Dict[str, Any] = {}

DEFAULTS["client"] = {"client": {"tracker": {}}}
DEFAULTS["experiment"] = {"experiment": {}}
DEFAULTS["task"] = {"task": {}}
DEFAULTS["study"] = {"study": {"objective": {}, "tuner": {}}}


def get(name: str) -> Dict[str, Any]:
    return copy.deepcopy(DEFAULTS[name])


DEFAULT_CLASS: Dict[str, Any] = {}

DEFAULT_CLASS["core"] = {
    "client": "ivory.core.client.Client",
    "tracker": "ivory.core.tracker.Tracker",
    "tuner": "ivory.core.tuner.Tuner",
    "experiment": "ivory.core.base.Experiment",
    "objective": "ivory.core.objective.Objective",
    "run": "ivory.core.run.Run",
    "task": "ivory.core.run.Task",
    "study": "ivory.core.run.Study",
    "dataset": "ivory.core.data.Dataset",
    "datasets": "ivory.core.data.Datasets",
    "results": "ivory.callbacks.results.Results",
    "metrics": "ivory.callbacks.metrics.Metrics",
    "monitor": "ivory.callbacks.monitor.Monitor",
    "early_stopping": "ivory.callbacks.early_stopping.EarlyStopping",
}

DEFAULT_CLASS["torch"] = {
    "run": "ivory.torch.run.Run",
    "dataset": "ivory.torch.data.Dataset",
    "results": "ivory.torch.results.Results",
    "metrics": "ivory.torch.metrics.Metrics",
    "trainer": "ivory.torch.trainer.Trainer",
}

DEFAULT_CLASS["tensorflow"] = {
    "run": "ivory.tensorflow.run.Run",
    "trainer": "ivory.tensorflow.trainer.Trainer",
}


DEFAULT_CLASS["sklearn"] = {
    "estimator": "ivory.sklearn.estimator.Estimator",
    "metrics": "ivory.sklearn.metrics.Metrics",
}


def update_class(params: Dict[str, Any], library: str = "core"):
    if 'library' in params:
        library = params.pop("library")
    for key, value in params.items():
        if value is None:
            value = {}
            params[key] = value
        if not isinstance(value, dict):
            continue
        if "library" in value:
            library = value.pop("library")
        for kind in ["class", "def", "call"]:
            if kind in value:
                attr = value[kind]
                break
        else:
            kind = "class" if key != "dataset" else "def"
            if key in DEFAULT_CLASS[library]:
                attr = DEFAULT_CLASS[library][key]
            elif key in DEFAULT_CLASS["core"]:
                attr = DEFAULT_CLASS["core"][key]
            else:
                attr = None
            if attr:
                value[kind] = attr
            else:
                params[key] = None
                break

        from ivory.core import instance

        attr = instance.get_attr(attr)
        if "__requires__" in dir(attr):
            requires = getattr(attr, "__requires__")
            for r in requires:
                if r not in value:
                    value[r] = {}
        update_class(value, library)



import gc
from typing import Callable, Dict, Iterable, Union

import numpy as np
import optuna
from optuna.trial import Trial

from ivory.callbacks.pruning import Pruning
from ivory.core import instance
from ivory.utils.range import Range


class Objective:
    def __init__(self, **suggests):
        self.suggests = {}
        self.update(**suggests)

    def update(self, args=None, **suggests):
        if args is None:
            args = {}
        args.update(suggests)
        for key, value in args.items():
            if callable(value):
                self.suggests[key] = value
            elif isinstance(value, str):
                self.suggests[key] = instance.get_attr(value)
            elif isinstance(key, str):
                self.suggests[key] = create_suggest({key: value})  # single parameter
            else:
                suggests = dict(zip(key, value))  # product parameters
                name = ".".join(suggests.keys())
                self.suggests[name] = create_suggest(suggests)

    def __repr__(self):
        class_name = self.__class__.__name__
        return f"{class_name}({list(self.suggests.keys())})"

    def __getitem__(self, name):
        return self.suggests[name]

    def __getattr__(self, name):
        return self.suggests[name]

    def __contains__(self, name):
        return name in self.suggests

    def __call__(
        self, suggest_name: str, create_run: Callable, has_pruning: bool
    ) -> Callable:
        suggest = self.suggests[suggest_name]

        def objective(trial: Trial):
            suggest(trial)
            run = create_run(trial.params)
            if run.tracking:
                run.tracking.set_tags(run.id, {"trial_number": trial.number})
                trial.set_user_attr("run_id", run.id)
            if has_pruning:
                run.set(pruning=Pruning(trial, run.monitor.metric))
            run.start("train")
            score = run.monitor.best_score
            if np.isnan(score):
                raise optuna.exceptions.TrialPruned("Best score is nan.")
            del run
            gc.collect()
            return score

        return objective


def create_suggest(params: Dict[str, Union[Callable, Iterable]]) -> Callable:
    """Creates a suggest function from a parameter dictionary."""
    suggests: Dict[str, Union[Callable, list]] = {}
    for key, value in params.items():
        if callable(value):
            suggests[key] = value
        elif isinstance(value, tuple):
            low, high = value
            type = "int" if isinstance(low, int) else "float"
            suggests[key] = [type, dict(low=low, high=high)]
        elif isinstance(value, range):
            low, high, step = value.start, value.stop - 1, value.step
            suggests[key] = ["int", dict(low=low, high=high, step=step)]
        elif isinstance(value, Range):
            low, high, step = value.start, value.stop, value.step
            if value.log:
                suggests[key] = ["float", dict(low=low, high=high, log=True)]
            elif value.is_integer:
                suggests[key] = ["int", dict(low=low, high=high, step=step)]
            else:
                if step == 1 and value.num == 0:
                    suggests[key] = ["float", dict(low=low, high=high)]
                else:
                    if value.num:
                        step = (high - low) / value.num  # type:ignore
                    args = dict(low=low, high=high, step=step)
                    suggests[key] = ["discrete_uniform", args]
        else:
            suggests[key] = ["categorical", dict(choices=list(value))]

    def suggest(trial: Trial, suggests=suggests):
        for key, value in suggests.items():
            if callable(value):
                value(trial)
            else:
                suggest = getattr(trial, "suggest_" + value[0])
                suggest(key, **value[1])

    return suggest



import importlib
import inspect
import re
import types
from functools import partial
from typing import Any, Callable, Dict, Iterable, Iterator, Tuple

from ivory.core.default import update_class


def get_attr(path):
    if not isinstance(path, str):
        return path
    if "." not in path:
        raise ValueError("module path not included")
    module_path, _, name = path.rpartition(".")
    module = importlib.import_module(module_path)
    return getattr(module, name)


def instantiate(params: Dict[str, Any], globals=None, kwargs=None):
    if globals is None:
        globals = {}
    else:
        globals = globals.copy()
    if kwargs is None:
        kwargs = {}
    return _instantiate(params, globals, kwargs)


def _instantiate(params: Dict[str, Any], globals, kwargs):
    if "class" in params:
        key = "class"
    elif "call" in params:
        key = "call"
    elif "def" in params:
        key = "def"
    else:
        raise ValueError("dict-key must include one of (class, call, def)")

    if key == "class" and "," in params[key]:
        paths = [path.strip() for path in params[key].split(",")]
        bases = [get_attr(path) for path in paths]
        name = paths[0].split(".")[-1]
        attr = types.new_class(name, bases)
    else:
        attr = get_attr(params[key])
    try:
        signature = inspect.signature(attr)
    except ValueError:
        pass
    else:
        parameters = signature.parameters
        for k, value in parameters.items():
            default = value.default
            if k in params and params[k] == "__default__":
                params[k] = default
    args = {k: v for k, v in params.items() if k != key}
    args = parse_value(args, globals, "")
    for k, value in args.items():
        if isinstance(params[k], str) and params[k].startswith("$"):
            if isinstance(value, (int, float, str, list, tuple, dict)):
                params[k] = value
    positional = []
    if "_" in args:
        positional = [args.pop("_")]
    if "__" in args:
        positional.extend(args.pop("__"))
    if key != "def":
        return attr(*positional, **args, **kwargs)
    else:
        if args or kwargs:
            return partial(attr, *positional, **args, **kwargs)
        else:
            return attr


def parse_value(value, globals, key: str):
    """
    Examples:
       >>> globals = {"a": 0, "b": [1, 2, 3]}
       >>> parse_value("$", globals, "a")
       0
       >>> parse_value("$.b", globals, "a")
       [1, 2, 3]
       >>> parse_value("$.b.1", globals, "a")
       2
       >>> parse_value("$.b.pop()", globals, "a")
       3
       >>> globals
       {'a': 0, 'b': [1, 2]}
    """
    if isinstance(value, dict):
        if "class" in value or "call" in value or "def" in value:
            obj = globals[key] = _instantiate(value, globals, {})
            return obj
        else:
            return {key: parse_value(value[key], globals, key) for key in value}
    elif isinstance(value, list):
        return [parse_value(v, globals, key) for v in value]
    elif value == "$":
        return globals[key]
    elif isinstance(value, str) and value.startswith("$."):
        value = value[2:]
        m = re.match(r"(.*)\.(\d+)$", value)
        if m:
            value, index = m.group(1), int(m.group(2))
        else:
            index = -1
        if "." in value:
            key, _, rest = value.partition(".")
            value = eval(f"globals[key].{rest}")
        else:
            value = globals[value]
        if index >= 0:
            value = value[index]
    return value


def create_base_instance(params: Dict[str, Any], name: str, source_name: str = ""):
    update_class(params)
    kwargs = dict(params=params, source_name=source_name)
    return instantiate(params[name], kwargs=kwargs)


def create_instance(params: Dict[str, Any], name: str, globals=None, **kwargs) -> Any:
    if globals is None:
        globals = {}
    globals.update(**kwargs)
    update_class(params)
    names = name.split(".")
    for name in names:
        params = params[name]
    return instantiate(params, globals)


def create_instances(params: Dict[str, Any], names: Iterable[str]) -> Iterator[Any]:
    globals: Dict[str, Any] = {}
    for name in names:
        instance = create_instance(params, name, globals)
        yield instance
        globals[name] = instance


def filter_params(func: Callable, **kwargs) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    params = {}
    kwargs_ = {}
    parameters = inspect.signature(func).parameters.keys()
    for key, value in kwargs.items():
        if key in parameters:
            kwargs_[key] = value
        else:
            params[key] = value
    return params, kwargs_



import torch

import ivory.core.run


class Run(ivory.core.run.Run):
    def save_instance(self, state_dict, path):
        torch.save(state_dict, path)

    def load_instance(self, path):
        return torch.load(path)



import os
import re
import subprocess
import sys

from setuptools import setup


def get_version(package: str) -> str:
    path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), package, "__init__.py"
    )
    with open(path, "r") as file:
        source = file.read()
    m = re.search(r'__version__ = ["\'](.+)["\']', source)
    if m:
        return m.group(1)
    else:
        return "0.0.0"


def get_packages(package):
    """Return root package and all sub-packages."""
    return [
        dirpath
        for dirpath, dirnames, filenames in os.walk(package)
        if os.path.exists(os.path.join(dirpath, "__init__.py"))
    ]


def check():
    def run(command):
        assert subprocess.run(command.split()).returncode == 0
        print(f"'{command}' --- OK")

    run("pycodestyle ivory")
    run("pyflakes ivory")
    run("mypy ivory")
    run("pycodestyle tests")
    run("pyflakes tests")
    run("mypy tests")


def publish():
    check()
    subprocess.run("python setup.py sdist bdist_wheel".split())
    subprocess.run("twine upload dist/*".split())
    version = get_version("ivory")
    subprocess.run(["git", "tag", "-a", f"v{version}", "-m", f"'Version {version}'"])
    subprocess.run(["git", "push", "origin", "--tags"])
    sys.exit(0)


if sys.argv[-1] == "publish":
    publish()

if sys.argv[-1] == "check":
    check()


long_description = "ML Tool."

setup(
    name="ivory",
    version=get_version("ivory"),
    description="ML Tool",
    long_description=long_description,
    url="https://ivory.daizutabi.net",
    author="daizutabi",
    author_email="daizutabi@gmail.com",
    license="MIT",
    packages=get_packages("ivory"),
    include_package_data=True,
    entry_points={"console_scripts": ["ivory = ivory.main:main"]},
    install_requires=[
        "mlflow",
        "optuna",
        "numpy",
        "pandas",
        "scikit-learn",
        "termcolor",
        "click",
        "logzero",
        "tqdm",
    ],
    classifiers=[
        "Environment :: Console",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



class IvoryError(Exception):
    """Base class for Ivory specific errors."""


class EarlyStopped(IvoryError):
    """Exception for early stopped runs.

    This error tells a trainer that the current `run` should be early stopped.
    """


class Pruned(IvoryError):
    """Exception for pruned runs.

    This error tells a trainer that the current `run` should  be pruned.
    """


class TestDataNotFoundError(IvoryError):
    """Exception when test data not found."""
